#include "auth-pam.h"
#include <qt5-log-i.h>
#include <QDateTime>
#include <QDebug>
#include <QTextCodec>

#define PAM_SERVICE_NAME "mate-screensaver"
#define MAX_TRIES 5

void no_fail_delay(int status, unsigned int delay, void *appdata_ptr)
{
}

AuthPam::AuthPam(QObject *parent)
    : QThread(parent)
{
    connect(this, &QThread::finished, [this]() {
        resetFlag();
    });
}

AuthPam::~AuthPam()
{
    if (!isFinished())
    {
        QThread::requestInterruption();
        m_waitCondition.wakeAll();
        m_reAuthCOndition.wakeAll();
        if (!QThread::wait(1000))
        {
            qWarning() << "pam auth thread can't stop,terminate it." << m_state;
            QThread::terminate();
            QThread::wait();
        }
    }
    //清理标志
    resetFlag();
}

void AuthPam::startAuthenticate(const QString &userName)
{
    if (m_state != AUTH_STATE_NOT_RUNNING)
    {
        KLOG_WARNING() << "start failed,is running";
        return;
    }
    m_authUserName = userName;
    m_state = AUTH_STATE_RUNNING;
    m_authRes = false;
    start();
}

void AuthPam::reAuthenticate()
{
    if (m_state != AUTH_STATE_WAIT_RE_AUTH)
    {
        KLOG_WARNING() << "current state is not wait reauth";
        return;
    }
    m_reAuthCOndition.wakeAll();
}

void AuthPam::response(const QString &str)
{
    if (m_state != AUTH_STATE_WAIT_RESPONSE)
    {
        KLOG_WARNING() << "current state is nor wait response";
        return;
    }
    m_conversationRep = {true, str};
    m_waitCondition.wakeAll();
    m_state = AUTH_STATE_RUNNING;
}

int AuthPam::conversation(int num_msg, const pam_message **msg, pam_response **resp, void *appdata_ptr)
{
    AuthPam *This = static_cast<AuthPam *>(appdata_ptr);
    struct pam_response *reply = NULL;
    bool res;
    int ret;
    int replies;

    reply = (struct pam_response *)calloc(num_msg, sizeof(*reply));
    if (reply == nullptr)
    {
        return PAM_CONV_ERR;
    }

    res = true;
    ret = PAM_SUCCESS;
    for (replies = 0; replies < num_msg && ret == PAM_SUCCESS; replies++)
    {
        if (This->isInterruptionRequested())
        {
            KLOG_INFO("thread interruption is set,skip msg [%s]", msg[replies]->msg ? msg[replies]->msg : "null");
            goto next_msg;
        }
        KLOG_INFO() << "conversation type[" << msg[replies]->msg_style << "]"
                     << " " << QString::fromLocal8Bit(msg[replies]->msg);

        switch (msg[replies]->msg_style)
        {
        case PAM_PROMPT_ECHO_ON:
        {
            /// PAM返回编码为系统设置的编码形式,需手动转换Qt使用的UTF-8
            emit This->showPrompt(QString::fromLocal8Bit(msg[replies]->msg), Kiran::PromptTypeQuestion);
            This->waitForResponse();
            if(!This->m_conversationRep.first && This->m_conversationRep.second.isEmpty() )
            {
                res =false;
            }
            reply[replies].resp_retcode = This->m_conversationRep.first ? PAM_SUCCESS : PAM_CONV_ERR;
            std::string str = This->m_conversationRep.second.toStdString();
            reply[replies].resp = new char[str.length() + 1];
            strcpy(reply[replies].resp, str.c_str());
            break;
        }
        case PAM_PROMPT_ECHO_OFF:
        {
            emit This->showPrompt(QString::fromLocal8Bit(msg[replies]->msg), Kiran::PromptTypeSecret);
            This->waitForResponse();
            if(!This->m_conversationRep.first && This->m_conversationRep.second.isEmpty() )
            {
                res =false;
            }
            reply[replies].resp_retcode = This->m_conversationRep.first ? PAM_SUCCESS : PAM_CONV_ERR;
            std::string str = This->m_conversationRep.second.toStdString();
            reply[replies].resp = new char[str.length() + 1];
            strcpy(reply[replies].resp, str.c_str());
            break;
        }
        case PAM_ERROR_MSG:
        {
            emit This->showMessage(QString::fromLocal8Bit(msg[replies]->msg), Kiran::MessageTypeError);
            break;
        }
        case PAM_TEXT_INFO:
        {
            emit This->showMessage(QString::fromLocal8Bit(msg[replies]->msg), Kiran::MessageTypeInfo);
            break;
        }
        default:
            break;
        }
    next_msg:
        if ( res )
        {
            reply[replies].resp_retcode = PAM_SUCCESS;
        }
        else
        {
            int i;
            for (i = 0; i <= replies; i++)
            {
                if (reply[i].resp != nullptr)
                    free(reply[i].resp);
            }
            free(reply);
            reply = nullptr;
            ret = PAM_CONV_AGAIN;
        }
    }
    *resp = reply;
    return ret;
}

void AuthPam::waitForResponse()
{
    m_waitMutex.lock();
    m_state = AUTH_STATE_WAIT_RESPONSE;
    m_waitCondition.wait(&m_waitMutex);
    m_waitMutex.unlock();
}

void AuthPam::start(QThread::Priority priority)
{
    QThread::start();
}

void AuthPam::run()
{
    const struct pam_conv conv = {
        &AuthPam::conversation,
        this};

    pam_handle_t *pamh = nullptr;
    int iRet = 0;

    KLOG_INFO() << "pam_start-->" << m_authUserName;
    iRet = pam_start(PAM_SERVICE_NAME, m_authUserName.toStdString().c_str(), &conv, &pamh);
    if (iRet != PAM_SUCCESS)
    {
        //goto clean
        KLOG_WARNING() << "pam_start failed," << pam_strerror(pamh, iRet);
        goto end;
    }
    pam_set_item(pamh,PAM_FAIL_DELAY,(void*)no_fail_delay);

    while (!QThread::isInterruptionRequested())
    {
        m_state = AUTH_STATE_RUNNING;
        iRet = pam_authenticate(pamh, 0);
        if (iRet != PAM_SUCCESS)
        {
            KLOG_WARNING() << "pam_authenticate failed," << pam_strerror(pamh, iRet);
            if( !QThread::isInterruptionRequested() )
            {
                emit authenticationComplete();
                waitForReAuthenticate();
                continue;
            }
        }
        else
        {
            m_authRes = true;
            emit authenticationComplete();
            break;
        }
    }

end:
    pam_end(pamh, iRet);
    m_authRes = iRet == PAM_SUCCESS;
    KLOG_INFO() << "pam auth thread finished"
                 << "pam_end" << iRet;
}

bool AuthPam::isAuthenticated()
{
    return m_authRes;
}

void AuthPam::waitForReAuthenticate()
{
    m_reAuthMutex.lock();
    m_state = AUTH_STATE_WAIT_RE_AUTH;
    m_reAuthCOndition.wait(&m_reAuthMutex);
    m_reAuthMutex.unlock();
}

AuthPam::AuthState AuthPam::state()
{
    return m_state;
}

void AuthPam::resetFlag()
{
    m_state = AUTH_STATE_NOT_RUNNING;
    m_waitMutex.unlock();
    m_waitCondition.wakeAll();
    m_reAuthMutex.unlock();
    m_reAuthCOndition.wakeAll();
    m_conversationRep = {false, ""};
}
